#include "mex.h"
#include "math.h"
#include "stdio.h"

#define TOL 1.0E-4

void mexFunction(int nlhs, mxArray *plhs[], 
    int nrhs, const mxArray *prhs[])

{
 int N, M;
 if (nrhs < 8)
	mexErrMsgTxt("The number of input arguments must be eight.");

 /* Find the dimensions of the input matrices; */
 N = mxGetN(prhs[0]);
 M = mxGetM(prhs[0]);
 
 if (N != mxGetM(prhs[5]))
	mexErrMsgTxt("Error: Dimensions must agree.");
  
 /* Create an mxArray for the output data */
 plhs[0] = mxCreateDoubleMatrix(N, 1, mxREAL);
 plhs[1] = mxCreateDoubleMatrix(N, 1, mxREAL);
 plhs[2] = mxCreateDoubleMatrix(N, 1, mxREAL);
 plhs[3] = mxCreateDoubleMatrix(N, 1, mxREAL);

 double *W_w, *W_x, *W_y, *W_z;   
 double *u_w, *u_x, *u_y, *u_z; 
 double *x_w, *x_x, *x_y, *x_z; 
 double *Energy;

 /* Retrieve the input data */
 W_w = mxGetPr(prhs[0]);
 W_x = mxGetPr(prhs[1]);
 W_y = mxGetPr(prhs[2]);
 W_z = mxGetPr(prhs[3]);
 u_w = mxGetPr(prhs[4]);
 u_x = mxGetPr(prhs[5]);
 u_y = mxGetPr(prhs[6]);
 u_z = mxGetPr(prhs[7]);

 /* Create a pointer to the output data */
 x_w = mxGetPr(plhs[0]);
 x_x = mxGetPr(plhs[1]);
 x_y = mxGetPr(plhs[2]);
 x_z = mxGetPr(plhs[3]);
 
 /* Create the old state vector. */
 double old_w[N], old_x[N], old_y[N], old_z[N];

 /* Put data in the output array */
 double v_w, v_x, v_y, v_z, Av, Error=1;
 double X1, X2, X3;
 int i, j, it, it_max;

 if (nrhs < 9)
	it_max = 100; 
 else
	it_max = mxGetScalar(prhs[8]);
 it_max++;
 
 /* Energy function */
 plhs[4] = mxCreateDoubleMatrix(N*it_max, 1, mxREAL);
 Energy = mxGetPr(plhs[4]);
 
 it=0;
 /* Compute the energy */
 Energy[it]=0;
 for (i=0; i<N; i++)
 {
   /* Store old and current state */
   old_w[i]=u_w[i];
   old_x[i]=u_x[i];
   old_y[i]=u_y[i];
   old_z[i]=u_z[i];
       
   x_w[i] = u_w[i];
   x_x[i] = u_x[i];
   x_y[i] = u_y[i];
   x_z[i] = u_z[i];
   
   /* Compute the potential of the i$th$ neuron */
   v_w=0;
   v_x=0;
   v_y=0;
   v_z=0;
   for (j=0; j<N; j++)
     {
       v_w = v_w + W_w[i+M*j]*u_w[j] - W_x[i+M*j]*u_x[j] - W_y[i+M*j]*u_y[j] - W_z[i+M*j]*u_z[j];
       v_x = v_x + W_y[i+M*j]*u_z[j] - W_z[i+M*j]*u_y[j] + W_w[i+M*j]*u_x[j] + W_x[i+M*j]*u_w[j];
       v_y = v_y + W_z[i+M*j]*u_x[j] - W_x[i+M*j]*u_z[j] + W_w[i+M*j]*u_y[j] + W_y[i+M*j]*u_w[j];
       v_z = v_z + W_x[i+M*j]*u_y[j] - W_y[i+M*j]*u_x[j] + W_w[i+M*j]*u_z[j] + W_z[i+M*j]*u_w[j];
     }
   Energy[it]=Energy[it]-(x_w[i]*v_w+x_x[i]*v_x+x_y[i]*v_y+x_z[i]*v_z)/2;
 }
 
 /* Iterate the network */  
 while ((it<it_max)&&(Error>TOL))
 {
   it=it+1;
   Error = 0;
   for (i=0; i<M; i++)
   {
     /* Compute the potential of the i$th$ neuron */
     v_w=0;
     v_x=0;
     v_y=0;
     v_z=0;
     for (j=0; j<N; j++)
     {
       v_w = v_w + W_w[i+M*j]*old_w[j] - W_x[i+M*j]*old_x[j] - W_y[i+M*j]*old_y[j] - W_z[i+M*j]*old_z[j];
       v_x = v_x + W_y[i+M*j]*old_z[j] - W_z[i+M*j]*old_y[j] + W_w[i+M*j]*old_x[j] + W_x[i+M*j]*old_w[j];
       v_y = v_y + W_z[i+M*j]*old_x[j] - W_x[i+M*j]*old_z[j] + W_w[i+M*j]*old_y[j] + W_y[i+M*j]*old_w[j];
       v_z = v_z + W_x[i+M*j]*old_y[j] - W_y[i+M*j]*old_x[j] + W_w[i+M*j]*old_z[j] + W_z[i+M*j]*old_w[j];
     }
     
     /* Compute the absolute value of the potential of the ith neuron; */
     Av = sqrt(v_w*v_w + v_x*v_x + v_y*v_y + v_z*v_z);
     /* Evaluate the function sigma */
     if (Av>TOL)
     {
       x_w[i] = v_w/Av;
       x_x[i] = v_x/Av;
       x_y[i] = v_y/Av;
       x_z[i] = v_z/Av;
     }
          
     /* Compute the error */
     Error = Error + sqrt((old_w[i] - x_w[i])*(old_w[i] - x_w[i]) + (old_x[i] - x_x[i])*(old_x[i] - x_x[i]) + (old_y[i] - x_y[i])*(old_y[i] - x_y[i]) + (old_z[i] - x_z[i])*(old_z[i] - x_z[i]));
   }
   
   /* Update the state of the network and compute the energy */
   
   Energy[it]=0;  
   for (i=0; i<N; i++) 
   {
      old_w[i] = x_w[i];
      old_x[i] = x_x[i];
      old_y[i] = x_y[i];
      old_z[i] = x_z[i];
      /* Compute the potential of the i$th$ neuron */
      v_w=0;
      v_x=0;
      v_y=0;
      v_z=0;
      for (j=0; j<N; j++)
      {
        v_w = v_w + W_w[i+M*j]*x_w[j] - W_x[i+M*j]*x_x[j] - W_y[i+M*j]*x_y[j] - W_z[i+M*j]*x_z[j];
        v_x = v_x + W_y[i+M*j]*x_z[j] - W_z[i+M*j]*x_y[j] + W_w[i+M*j]*x_x[j] + W_x[i+M*j]*x_w[j];
        v_y = v_y + W_z[i+M*j]*x_x[j] - W_x[i+M*j]*x_z[j] + W_w[i+M*j]*x_y[j] + W_y[i+M*j]*x_w[j];
        v_z = v_z + W_x[i+M*j]*x_y[j] - W_y[i+M*j]*x_x[j] + W_w[i+M*j]*x_z[j] + W_z[i+M*j]*x_w[j];
      }
      Energy[it]=Energy[it]-(x_w[i]*v_w+x_x[i]*v_x+x_y[i]*v_y+x_z[i]*v_z)/2;  
   }
 }
 
 if (it<it_max)
   printf("The parallel CV-QHNN model reached a stationary state with %d iterations.\n",it-1);
 else
 {
     it=it-1;
     printf("The parallel CV-QHNN failed to reach a stationary state in %d iterations.\n",it);
 }
 plhs[5] = mxCreateDoubleScalar((double) it);
}
