#include "mex.h"
#include "math.h"
#include "stdio.h"

#define TOL 1.0E-4
#define PI 3.14159265358979323846

double Phi, Psi, Theta;

/* Get the angle-phase representation of a quaternion */
void AnglePhase(double a, double b, double c, double d)
{
  double Av2;
  
  Av2 = a*a+b*b+c*c+d*d;
  Psi = -asin(2*(b*c - a*d)/Av2)/2;
  if (fabs(fabs(Psi)-PI/4)<TOL)
  {
    Theta=0;
    Phi=atan2(2*(-c*d+a*b),a*a-b*b-c*c+d*d)/2;  
  }
  else
  {
    Phi=atan2(2*(c*d+a*b),a*a-b*b+c*c-d*d)/2;
    Theta=atan2(2*(b*d+a*c),a*a+b*b-c*c-d*d)/2;
  }
  Av2 = cos(Phi)*cos(Psi)*cos(Theta) + sin(Phi)*sin(Psi)*sin(Theta);
  if (a*Av2<0)
  {
    if (Phi < 0)
    {
	Phi = Phi + PI;
    }
    else
    {
      Phi = Phi - PI;
    }
  }

}

void mexFunction(int nlhs, mxArray *plhs[], 
    int nrhs, const mxArray *prhs[])

{
 int N, M;
 if (nrhs < 12)
	mexErrMsgTxt("The number of input arguments must be twelve.");

 /* Find the dimensions of the input matrices; */
 N = mxGetN(prhs[0]);
 M = mxGetM(prhs[0]);
 
 if (N != mxGetM(prhs[5]))
	mexErrMsgTxt("Error: Dimensions must agree.");
  
 /* Create an mxArray for the output data */
 plhs[0] = mxCreateDoubleMatrix(N, 1, mxREAL);
 plhs[1] = mxCreateDoubleMatrix(N, 1, mxREAL);
 plhs[2] = mxCreateDoubleMatrix(N, 1, mxREAL);
 plhs[3] = mxCreateDoubleMatrix(N, 1, mxREAL);

 double *W_w, *W_x, *W_y, *W_z;   
 double *u_w, *u_x, *u_y, *u_z; 
 double *x_w, *x_x, *x_y, *x_z;
 double K1, K2, K3;
 double *Energy;
 
 /* Retrieve the input data */
 W_w = mxGetPr(prhs[0]);
 W_x = mxGetPr(prhs[1]);
 W_y = mxGetPr(prhs[2]);
 W_z = mxGetPr(prhs[3]);
 u_w = mxGetPr(prhs[4]);
 u_x = mxGetPr(prhs[5]);
 u_y = mxGetPr(prhs[6]);
 u_z = mxGetPr(prhs[7]);
 K1 = mxGetScalar(prhs[8]);
 K2 = mxGetScalar(prhs[9]);
 K3 = mxGetScalar(prhs[10]);

 /* Create a pointer to the output data */
 x_w = mxGetPr(plhs[0]);
 x_x = mxGetPr(plhs[1]);
 x_y = mxGetPr(plhs[2]);
 x_z = mxGetPr(plhs[3]);

 /* Create the old state vector. */
 double old_w[N], old_x[N], old_y[N], old_z[N];
 
 /* Put data in the output array */
 double v_w, v_x, v_y, v_z; 
 double A, B, Av, Error=1;
 int i, j, r, it, it_max, itE;
 int f1, f2, f3;
 
 double Phi_x[N], Psi_x[N], Theta_x[N];
 double Psi0, Theta0, Phi0;
 Phi0 = 2*PI/(K1); 
 Psi0 = PI/(2*K2);
 Theta0 = PI/(K3);

 if (nrhs < 11)
	it_max = 100; 
 else
	it_max = mxGetScalar(prhs[11]);
 it_max++;
 
 /* Energy function */
 plhs[4] = mxCreateDoubleMatrix(3*N*it_max, 1, mxREAL);
 Energy = mxGetPr(plhs[4]);
  
 it=0;
  /* Compute the energy */
 itE=0;
 Energy[itE]=0;
 for (i=0; i<N; i++)
 {
   /* Store old and current state */
   old_w[i]=u_w[i];
   old_x[i]=u_x[i];
   old_y[i]=u_y[i];
   old_z[i]=u_z[i];
     
   x_w[i] = u_w[i];
   x_x[i] = u_x[i];
   x_y[i] = u_y[i];
   x_z[i] = u_z[i];
   
   AnglePhase(x_w[i], x_x[i], x_y[i], x_z[i]);
   Phi_x[i] = Phi;
   Psi_x[i] = Psi;
   Theta_x[i] = Theta;  
   
   /* Compute the potential of the i$th$ neuron */
   v_w=0;
   v_x=0;
   v_y=0;
   v_z=0;
   for (j=0; j<N; j++)
     {
       v_w = v_w + W_w[i+M*j]*u_w[j] - W_x[i+M*j]*u_x[j] - W_y[i+M*j]*u_y[j] - W_z[i+M*j]*u_z[j];
       v_x = v_x + W_y[i+M*j]*u_z[j] - W_z[i+M*j]*u_y[j] + W_w[i+M*j]*u_x[j] + W_x[i+M*j]*u_w[j];
       v_y = v_y + W_z[i+M*j]*u_x[j] - W_x[i+M*j]*u_z[j] + W_w[i+M*j]*u_y[j] + W_y[i+M*j]*u_w[j];
       v_z = v_z + W_x[i+M*j]*u_y[j] - W_y[i+M*j]*u_x[j] + W_w[i+M*j]*u_z[j] + W_z[i+M*j]*u_w[j];
     }
   Energy[itE]=Energy[itE]-(x_w[i]*v_w+x_x[i]*v_x+x_y[i]*v_y+x_z[i]*v_z)/2;
 }
 
 /* Iterate the network */
 while ((it<it_max)&&(Error>TOL))
 {
   it=it+1;
   Error = 0;     
   for (r=0; r<3; r++)
   {
    /* Update all the neurons of the network */   
    for (i=0; i<N; i++)  
    {
     /* Compute the potential of the i$th$ neuron */
     v_w=0;
     v_x=0;
     v_y=0;
     v_z=0;
     for (j=0; j<N; j++)
     {
       v_w = v_w + W_w[i+M*j]*old_w[j] - W_x[i+M*j]*old_x[j] - W_y[i+M*j]*old_y[j] - W_z[i+M*j]*old_z[j];
       v_x = v_x + W_y[i+M*j]*old_z[j] - W_z[i+M*j]*old_y[j] + W_w[i+M*j]*old_x[j] + W_x[i+M*j]*old_w[j];
       v_y = v_y + W_z[i+M*j]*old_x[j] - W_x[i+M*j]*old_z[j] + W_w[i+M*j]*old_y[j] + W_y[i+M*j]*old_w[j];
       v_z = v_z + W_x[i+M*j]*old_y[j] - W_y[i+M*j]*old_x[j] + W_w[i+M*j]*old_z[j] + W_z[i+M*j]*old_w[j];
     }
     
     /* Compute the square of the absolute value of the activation potential */
     Av = sqrt(v_w*v_w + v_x*v_x + v_y*v_y + v_z*v_z);
     
     /* Evaluate the activation function if the potental activation is non-zero */
     if (Av>TOL)
     {
       /* Compute the components of the angle-phase representation of the activation potential */
       AnglePhase(v_w, v_x, v_y, v_z);
       /* printf("x[%d] = [%2.2f, %2.2f, %2.2f, %2.2f] (%2.2f, %2.2f, %2.2f)", i, old_w, old_x, old_y, old_z, Phi_x[i], Psi_x[i], Theta_x[i]);
       printf("-> [%2.2f, %2.2f, %2.2f, %2.2f] (%2.2f, %2.2f, %2.2f)", v_w, v_x, v_y, v_z, Phi, Psi, Theta); */
       
       if (fabs(fabs(Psi)-PI/4)>TOL)
       {
	if (r == 0)
	{
	  f1 = floor((PI+Phi)/Phi0);
	  Phi   = 0.5*(-2*PI+Phi0*(2*f1+1));
          
	  Phi_x[i] = Phi;
          Psi = Psi_x[i];
	  Theta = Theta_x[i];
	  /* printf("Update angle phi.  \n"); */ 
	}
	
	if (r == 1)
        {
          f2 = floor((0.25*PI+Psi)/Psi0);
	  Psi = 0.5*(-0.5*PI+Psi0*(2*f2+1));
	  
          Phi = Phi_x[i];
	  Psi_x[i] = Psi;
          Theta = Theta_x[i];
	  /* printf("Update angle psi \n"); */
	}

        if (r == 2)
        {
	  f3  = floor((0.5*PI+Theta)/Theta0);
          Theta = 0.5*(-PI+Theta0*(2*f3+1));
	  
          Phi = Phi_x[i];
	  Psi = Psi_x[i];
	  Theta_x[i]=Theta;
          /* printf("Update angle theta \n"); */
        }
        /* printf("Phase angles: phi = %2.4f, psi = %2.4f, theta = %2.4f.\n",Phi,Psi,Theta); */
	/* Update the current state x */     
	x_w[i] = cos(Phi)*cos(Psi)*cos(Theta) + sin(Phi)*sin(Psi)*sin(Theta);
	x_x[i] = sin(Phi)*cos(Psi)*cos(Theta) - cos(Phi)*sin(Psi)*sin(Theta);
	x_y[i] = cos(Phi)*cos(Psi)*sin(Theta) - sin(Phi)*sin(Psi)*cos(Theta);
	x_z[i] = cos(Phi)*sin(Psi)*cos(Theta) + sin(Phi)*cos(Psi)*sin(Theta);
       }
       // printf("=>  [%2.2f, %2.2f, %2.2f, %2.2f] (%2.2f, %2.2f, %2.2f) \n", x_w[i], x_x[i], x_y[i], x_z[i], Phi_x[i], Psi_x[i], Theta_x[i]);
     }
     
     /* Compute the error */
     Error = Error + sqrt((old_w[i] - x_w[i])*(old_w[i] - x_w[i]) + (old_x[i] - x_x[i])*(old_x[i] - x_x[i]) + (old_y[i] - x_y[i])*(old_y[i] - x_y[i]) + (old_z[i] - x_z[i])*(old_z[i] - x_z[i]));
    }
   
    /* Update the state of the network and compute the energy */
    itE++;
    Energy[itE]=0;  
    for (i=0; i<N; i++) 
    {
      old_w[i] = x_w[i];
      old_x[i] = x_x[i];
      old_y[i] = x_y[i];
      old_z[i] = x_z[i];
      /* Compute the potential of the i$th$ neuron */
      v_w=0;
      v_x=0;
      v_y=0;
      v_z=0;
      for (j=0; j<N; j++)
      {
        v_w = v_w + W_w[i+M*j]*x_w[j] - W_x[i+M*j]*x_x[j] - W_y[i+M*j]*x_y[j] - W_z[i+M*j]*x_z[j];
        v_x = v_x + W_y[i+M*j]*x_z[j] - W_z[i+M*j]*x_y[j] + W_w[i+M*j]*x_x[j] + W_x[i+M*j]*x_w[j];
        v_y = v_y + W_z[i+M*j]*x_x[j] - W_x[i+M*j]*x_z[j] + W_w[i+M*j]*x_y[j] + W_y[i+M*j]*x_w[j];
        v_z = v_z + W_x[i+M*j]*x_y[j] - W_y[i+M*j]*x_x[j] + W_w[i+M*j]*x_z[j] + W_z[i+M*j]*x_w[j];
      }
      Energy[itE]=Energy[itE]-(x_w[i]*v_w+x_x[i]*v_x+x_y[i]*v_y+x_z[i]*v_z)/2;  
     }
   }
 } 
 
 if (it<it_max)
   printf("The parallel MV-QHNN of Minemoto reached a stationary state with %d iterations.\n",it);
 else
 {
     it=it-1;
     printf("The parallel MV-QHNN of Minemoto failed to reach a stationary state in %d iterations.\n",it);
 }
 plhs[5] = mxCreateDoubleScalar((double) it);
}

